from .curve_generator import _tpl_curve_generator


class _tpl_curve_generator_correlated_data(_tpl_curve_generator):
  _identifiers = (
    "correlated data",
    "lines and points",
    "lines + points",
    "pointslines"
  )

  def _get_draw(self):
    color = "=%s" % self.curve.color.get_value() if self.curve.color.get_value() is not None else ""
    return "draw%s," % color

  def _get_marks(self):
    return "% marks,"



from collections.abc import Iterable
import csv
import subprocess

from .label import _tpl_label
from .curve import _tpl_curve
from .legend import _tpl_legend
from .strutils import _savetext, _compactify
from .logmodes import _tpl_logmodes
from .ticklabel import _tpl_ticklabel
from .cycle_list import _tpl_cycle_list


class figure:
  PDF_COMPILE_CMD = "pdflatex -halt-on-error -interaction nonstopmode TEXFILE | grep '^!.*' -A200 --color=always"

  def __init__(self, figsize: tuple = None):
    if figsize is None:
      figsize = (8, 5)
    self.__figsize: tuple = figsize
    self.__curves: list[_tpl_curve] = []
    self.__xlabel: _tpl_label = _tpl_label(token="xlabel")
    self.__ylabel: _tpl_label = _tpl_label(token="ylabel")
    self.__xticklabel: _tpl_ticklabel = _tpl_ticklabel("x")
    self.__yticklabel: _tpl_ticklabel = _tpl_ticklabel("y")
    self.__legend: _tpl_legend = _tpl_legend()
    self.__type: str = "xy"
    self.__logmodes: _tpl_logmodes = _tpl_logmodes()
    self.__additional_axis_cmds: list[str] = []
    self.__cycle_list: _tpl_cycle_list = _tpl_cycle_list("multi")

  def plot(self, x: Iterable, y: Iterable, xerror: Iterable = None, yerror: Iterable = None,
           *args, **kwargs):
    if len(x) != len(y):
      print("x and y must be of same length (%d != %d)" % (len(x), len(y)))
      return
    self.__curves.append(_tpl_curve(x, y, xerror, yerror, *args, **kwargs))

  def plotcsv(self, filename: str, xcolname: str, ycolname: str, xerrcolname: str = None, yerrcolname: str = None,
              *args, **kwargs):
    with open(filename, "r") as f:
      csvreader = csv.reader(f)
      data = dict()
      for d in list(map(list, zip(*([x for x in csvreader if x != []])))):
        data[d[0]] = d[1:]

      def error(axis: str, v: str):
        print("Cannot find %s-key \"%s\" in csv file %s" % (axis, v, filename))
        print("Found the following data:")
        print(data)
      if xcolname not in data.keys():
        error("x", xcolname)
        return
      if ycolname not in data.keys():
        error("y", ycolname)
        return
      if xerrcolname is not None and xerrcolname not in data.keys():
        error("x error", xerrcolname)
        return
      if yerrcolname is not None and yerrcolname not in data.keys():
        error("y error", yerrcolname)
        return
      xerr = data[xerrcolname] if xerrcolname in data.keys() else None
      yerr = data[yerrcolname] if yerrcolname in data.keys() else None

      self.__curves.append(_tpl_curve(data[xcolname], data[ycolname], xerr, yerr, *args, **kwargs))

  def savefig(self, filename: str):
    DEFAULT_SUFFIX = "tex"
    splitfile = filename.split(".")
    suffix = splitfile[-1] if len(splitfile) != 1 else None
    if suffix is None:
      suffix = DEFAULT_SUFFIX
      filename += "." + DEFAULT_SUFFIX
      print(f"No file-suffix specified. Assuming {DEFAULT_SUFFIX}.")
    if suffix.lower() == "tex":
      texstr = self.__create_tex()
      _savetext(filename, texstr)
    elif suffix.lower() == "tikz":
      tikzstr = self.__create_tikz()
      _savetext(filename, tikzstr)
    elif suffix.lower() == "pdf":
      texfilename = ".".join(filename.split(".")[:-1]) + ".tex"
      texstr = self.__create_tex()
      _savetext(texfilename, texstr)
      subprocess.run(self.__class__.PDF_COMPILE_CMD.replace("TEXFILE", texfilename).split())
    else:
      print("File type \"%s\" not supported!" % suffix)

  def set_xlabel(self, label: str):
    self.__xlabel.text.set_value(label)

  def set_ylabel(self, label: str):
    self.__ylabel.text.set_value(label)

  def set_xprecision(self, precision: int):
    self.__xticklabel._set_precision(precision)
    self.set_fixed_x_ticks(True)

  def set_yprecision(self, precision: int):
    self.__yticklabel._set_precision(precision)
    self.set_fixed_y_ticks(True)

  def set_fixed_x_ticks(self, fixed: bool = True):
    self.__xticklabel._set_fixed(fixed)
    self.__xticklabel._set_zerofill(fixed)

  def set_fixed_y_ticks(self, fixed: bool = True):
    self.__yticklabel._set_fixed(fixed)
    self.__yticklabel._set_zerofill(fixed)

  def set_scaled_x_ticks(self, scaled_ticks: bool = True):
    self.__xticklabel._set_scaled(scaled_ticks)

  def set_scaled_y_ticks(self, scaled_ticks: bool = True):
    self.__yticklabel._set_scaled(scaled_ticks)

  def set_x_ticks(self, data: list[float] = None):
    self.__xticklabel._set_ticks(data)

  def set_y_ticks(self, data: list[float] = None):
    self.__yticklabel._set_ticks(data)

  def set_x_ticks_from_data(self, use_data: bool = True):
    self.__xticklabel._set_ticks_use_data(use_data)

  def set_y_ticks_from_data(self, use_data: bool = True):
    self.__yticklabel._set_ticks_use_data(use_data)

  def set_x_tick_labels(self, labels: list[str] = None):
    self.__xticklabel._set_tick_labels(labels)

  def set_y_tick_labels(self, labels: list[str] = None):
    self.__yticklabel._set_tick_labels(labels)

  def set_legend_position(self, pos: str):
    self.__legend.pos.set_value(pos)

  def set_plottype(self, type: str = 'xy', idx: int = -1):
    self.__curves[idx]._set_plottype(type)

  def set_logmode(self, mode: str = None):
    self.__logmodes.set_mode(mode)

  def add_axis_command(self, cmd: str):
    self.__additional_axis_cmds.append(cmd)

  def repeat_style(self, ntimes: int = None):
    self.__cycle_list.repeat_style(ntimes)

  def __create_tex(self):
    preable_cmd = """\\documentclass[tikz]{standalone}\n
\\usepackage{pgfplots}
\\pgfplotsset{compat=newest}\n
\\begin{document}\n\n"""

    tikz_cmd = self.__create_tikz(True)

    end_cmd = "\n\\end{document}"

    return preable_cmd + tikz_cmd + end_cmd

  def __create_tikz(self, include_begintikz: bool = True):
    begintikz_cmd = "\\begin{tikzpicture}\n" if include_begintikz else ""

    body_cmd = f"""\\begin{{axis}}[
  label style={{
    font=\\large
  }},
  {self.__xlabel.text}
  {self.__ylabel.text}
  tick label style={{
    font=\\large
  }},
  {self.__xticklabel}
  {self.__yticklabel}
  {self.__legend.pos}
  legend cell align={{left}},
  grid=minor,
  grid style=dotted,
  scale only axis,
  {self.__logmodes.x}
  {self.__logmodes.y}
  {self.__cycle_list}"""
    body_cmd += ''.join(["\n  " + x + "," for x in self.__additional_axis_cmds])
    body_cmd += "\n]\n\n"

    plot_cmd = [c._get_addplot_code() for c in self.__curves]

    post_cmd = "\n\\end{axis}\n"

    endtikz_cmd = "\\end{tikzpicture}\n" if include_begintikz else ""

    return begintikz_cmd \
      + body_cmd \
      + "\n".join([_compactify(pc) for pc in plot_cmd]) \
      + post_cmd \
      + endtikz_cmd



def _compactify(s: str, level: int = 2):
  if level == 0:
    return s
  s = s.strip()
  if level == 1:
    return s
  s = "\n".join([x for x in s.split("\n") if x.strip() != ""])
  if level == 2:
    return s
  s = "\n".join([x for x in s.split("\n") if not x.strip().startswith("%")])
  if level == 3:
    return s


def _savetext(filename: str, text: str):
  with open(filename, "w") as f:
    f.writelines(text)


def _escape_characters(s: str, chars_to_escape: list[str] = None):
  if s is None:
    return s
  if chars_to_escape is None:
    chars_to_escape = ['%', '#']
  for ch in chars_to_escape:
    s = s.replace(ch, "\\" + ch)
  return s



from .curve_generator import _tpl_curve_generator


class _tpl_curve_generator_ybar(_tpl_curve_generator):
  _identifiers = (
    "bar",
    "ybar",
    "hist",
    "histogram"
  )

  def _get_fill(self):
    color = "=%s" % self.curve.color.get_value() if self.curve.color.get_value() is not None else ""
    return "fill%s,\n  fill opacity=0.3," % color

  def _get_draw(self):
    color = "=%s" % self.curve.color.get_value() if self.curve.color.get_value() is not None else ""
    return "draw%s,\n  draw opacity=0.7," % color

  def _get_errdraw(self):
    return self._get_draw()

  def _get_type(self):
    return "ybar,"

  def _get_marks(self):
    return "no marks,"

  def _get_legend_type(self):
    return "ybar legend,"


class _tpl_curve_generator_ybarinterval(_tpl_curve_generator_ybar):
  _identifiers = (
    "interval",
    "ybar interval"
  )

  def _get_type(self):
      return "ybar interval,"



from ..curve_data import _tpl_curve_data


class _tpl_curve_generator:
  _identifiers = (
    "xy",
    "only points",
    "scatter"
  )

  def __init__(self, c: _tpl_curve_data):
    self.curve = c

  def __create_table(self):
    table = f"\n  {'x':>8}, {'y':>8}"
    if self.curve.xerror is None and self.curve.yerror is None:
      table += "\n"
      table += "\n".join([f"  {cx:8}, {cy:8}" for cx, cy in zip(self.curve.x, self.curve.y)])
    elif self.curve.xerror is not None and self.curve.yerror is None:
      table += f", {'dx':>8}\n"
      table += "\n".join([
        f"  {cx:8}, {cy:8}, {cdx:8}"
        for cx, cy, cdx in zip(self.curve.x, self.curve.y, self.curve.xerror)
      ])
    elif self.curve.xerror is None and self.curve.yerror is not None:
      table += f", {'dy':>8}\n"
      table += "\n".join([
        f"  {cx:8}, {cy:8}, {cdy:8}"
        for cx, cy, cdy in zip(self.curve.x, self.curve.y, self.curve.yerror)
      ])
    elif self.curve.xerror is not None and self.curve.yerror is not None:
      table += f", {'dx':>8}, {'dy':>8}\n"
      table += "\n".join([
        f"  {cx:8}, {cy:8}, {cdx:8}, {cdy:8}"
        for cx, cy, cdx, cdy in zip(self.curve.x, self.curve.y, self.curve.xerror, self.curve.yerror)
      ])
    table += "\n"
    return table

  def _get_fill(self):
    return "% fill"

  def _get_draw(self):
    return "% draw"

  def _get_errdraw(self):
    return "% draw"

  def _get_type(self):
    return "% type (ybar, ybar interval)"

  def _get_marks(self):
    return "only marks,"

  def _get_legend_type(self):
    return "% legend type (ybar legend)"

  def _get_errorstr(self, draw: bool = False):
    xerrorstr = "x dir=both,\n  x explicit," if self.curve.xerror is not None else "% x error options"
    yerrorstr = "y dir=both,\n  y explicit," if self.curve.yerror is not None else "% y error options"

    if self.curve.xerror is not None or self.curve.yerror is not None:
      errormarkstr = f"""error mark options={{
    rotate=90,
    mark size=4pt,
    {self._get_errdraw()}
  }},"""
    else:
      errormarkstr = "  % error mark options"

    return '\n  '.join([xerrorstr, yerrorstr, errormarkstr])

  def _get_addplot_code(self):
    return f"""\\addplot+[
  {self.curve.color}
  {self._get_fill()}
  {self._get_draw()}
  {self._get_marks()}
  {"error bars/.cd," if self.curve.xerror is not None or self.curve.yerror is not None else "% error options"}
  {self._get_errorstr()}
  {"/tikz/.cd," if self.curve.xerror is not None or self.curve.yerror is not None else "% end error options"}
  {self._get_type()}
  {self._get_legend_type()}
  {self.curve.opacity}
  {"forget plot," if self.curve.label.get_value() is None else "% forget plot"}
] table [
  x=x,
  {"x error=dx," if self.curve.xerror is not None else "% x error"}
  y=y,
  {"y error=dy," if self.curve.yerror is not None else "% y error"}
  col sep=comma
] {{{self.__create_table()}}};
{self.curve.label}
"""



from .attribute import _tpl_attribute_variable


class _tpl_cycle_list:
  def __init__(self, type="multi"):
    self.value = _tpl_attribute_variable(token="cycle %s list" % type)

  def __str__(self):
    return str(self.value)

  def repeat_style(self, times: int):
    if times is None:
      self.value.set_value(None)
    else:
      self.value.set_value("{color\\nextlist [%u of]mark list}" % times)



from .attribute import _tpl_attribute_variable


class _tpl_logmodes:
  def __init__(self):
    self.x = _tpl_attribute_variable(token='xmode')
    self.y = _tpl_attribute_variable(token='ymode')

  def set_mode(self, mode: str):
    if mode == 'linear' or mode == '' or mode is None:
      self.x.set_value(None)
      self.y.set_value(None)
    if 'x' in mode:
      self.x.set_value('log')
    if 'y' in mode:
      self.y.set_value('log')



from .curve_generator import _tpl_curve_generator
from .ybar import _tpl_curve_generator_ybar, _tpl_curve_generator_ybarinterval
from .lines import _tpl_curve_generator_lines
from .correlated_data import _tpl_curve_generator_correlated_data


__subclasses = (
  _tpl_curve_generator,
  _tpl_curve_generator_ybar,
  _tpl_curve_generator_ybarinterval,
  _tpl_curve_generator_lines,
  _tpl_curve_generator_correlated_data
)


def _get_generator(plottype: str):
  for sc in __subclasses:
    for identifier in sc._identifiers:
      if plottype.lower() == identifier.lower():
        return sc
  print(f"Cannot identify plottype \"{type}\". Keeping previous type (default: xy).")
  return None



from .strutils import _escape_characters


class _tpl_attribute:
  def __init__(self, token: str):
    if token is None:
      print("texplotlibx/attribute: NULL token not allowed!")
    self._token = token
    self._value = None

  def get_value(self):
    if self._value is None:
      return self._value
    elif type(self._value) is str:
      return _escape_characters(self._value)
    elif type(self._value) is bool:
      return self._value
    else:
      return "%g" % self._value

  def __str__(self):
    return f"% {self._token}"

  def set_value(self, value):
    self._value = value


class _tpl_attribute_variable(_tpl_attribute):
  def __init__(self, token: str, value=None):
    super().__init__(token)
    self.set_value(value)

  def __str__(self):
    v = self.get_value()
    if v is None:
      return super().__str__()
    else:
      if type(v) is bool:
        v = str(v).lower()
      return self._token + "=" + v + ","


class _tpl_attribute_flag(_tpl_attribute):
  def __init__(self, token: str, enable: bool = False):
    super().__init__(token)
    if enable is None:
      print(f"Cannot set state of flag \"{token}\" to None. Defaulting to False.")
      enable = False
    self.set_value(enable)

  def set_value(self, value: bool = True):
    super().set_value(value)

  def __str__(self):
    v = self.get_value()
    if v is None:
      print(f"texplotlibx/attribute: Unknown if statement \"{self._token}\" should be set or not. " +
            "Choosing to disable flag.")
    elif v:
      return f"{self._token},"
    else:
      return super().__str__()


class _tpl_attribute_function(_tpl_attribute):
  def __init__(self, token: str, value=None, suppress_semicolon=False):
    super().__init__(token)
    self.set_value(value)
    self.suppress_semicolon = suppress_semicolon

  def __str__(self):
    v = self.get_value()
    if v is None:
      return super().__str__()
    else:
      semicolon = "" if self.suppress_semicolon else ";"
      return f"\\{self._token}{{{str(v)}}}" + semicolon



from .curve_generator import _tpl_curve_generator


class _tpl_curve_generator_lines(_tpl_curve_generator):
  _identifiers = (
    "lines",
    "line",
    "continuous"
  )

  def _get_draw(self):
    color = "=%s" % self.curve.color.get_value() if self.curve.color.get_value() is not None else ""
    return "draw%s," % color

  def _get_marks(self):
    return "no marks,"



from .attribute import _tpl_attribute_variable, _tpl_attribute_flag


class _tpl_ticklabel:
  def __init__(self, axis: str):
    if axis not in ('x', 'y'):
      print(f'Cannot interpret axis \"{axis}\" for tick label')
      return None
    self.__axis = axis
    self.__show = False
    self.__precision = _tpl_attribute_variable(token='precision')
    self.__fixed = _tpl_attribute_flag(token='fixed')
    self.__zerofill = _tpl_attribute_flag(token='fixed zerofill')
    self.__fontsize = _tpl_attribute_variable(token='font')
    self.__scaled_ticks = _tpl_attribute_variable(token=f'scaled {axis} ticks')
    self.__ticks = _tpl_attribute_variable(token=f'{axis}tick')
    self.__tick_labels = _tpl_attribute_variable(token=f'{axis}ticklabels')

  def _set_precision(self, prec: int):
    self.__precision.set_value(prec)
    self.__show = True

  def _set_fixed(self, fixed: bool = True):
    self.__fixed.set_value(fixed)
    self.__show = True

  def _set_zerofill(self, zerofill: bool = True):
    self.__zerofill.set_value(zerofill)
    self.__show = True

  def _set_fontsize(self, fontsize: str = '\\large'):
    self.__fontsize.set_value(fontsize)
    self.__show = True

  def _set_scaled(self, scaled_ticks: bool = True):
    self.__scaled_ticks.set_value(scaled_ticks)

  def _set_ticks(self, data: list[float] = None):
    if data is not None:
      lst = '{' + ','.join(map(str, data)) + '}'
      self.__ticks.set_value(lst)
      self.__show = True
    else:
      self.__ticks.set_value(None)

  def _set_ticks_use_data(self, use_data: bool = False):
    if use_data:
      self.__ticks.set_value('data')
    else:
      self.__ticks.set_value(None)
    self.__show = True

  def _set_tick_labels(self, labels: list[str] = None):
    if labels is not None:
      for label in labels:
        if label[0] != '{' or label[-1] != '}':
          label = '{' + label + '}'
      lst = '{' + ','.join(labels) + '}'
      self.__tick_labels.set_value(lst)
      self.__show = True
    else:
      self.__tick_labels.set_value(None)

    {self.__tick_labels}

  def __str__(self):
    tls = f"""{self.__axis} tick label style={{
    /pgf/number format/.cd,
    {self.__fixed}
    {self.__zerofill}
    {self.__precision}
    {self.__fontsize}
    {self.__ticks}
    /tikz/.cd
  }},""" if self.__show else f"% {self.__axis} tick label style"
    return tls + f"\n  {self.__scaled_ticks}"



from collections.abc import Iterable, Callable
from .attribute import _tpl_attribute_function, _tpl_attribute_variable
from .curve_label import _tpl_curve_label

class _tpl_curve_data:
  __counter: int = 0

  @staticmethod
  def __sort(x: Iterable, y: Iterable, xerr: Iterable = None, yerr: Iterable = None):
    def __subsort(m: Iterable):
      return zip(*sorted(zip(*m)))
    if xerr is None and yerr is None:
      x, y = __subsort((x, y))
    elif xerr is None and yerr is not None:
      x, y, yerr = __subsort((x, y, yerr))
    elif xerr is not None and yerr is None:
      x, y, xerr = __subsort((x, y, xerr))
    else:
      x, y, xerr, yerr = __subsort((x, y, xerr, yerr))
    return x, y, xerr, yerr

  def __init__(self, x: Iterable, y: Iterable, xerror: Iterable = None, yerror: Iterable = None,
               label: str = None, color: str = None, opacity: str = None, yfunc: Callable = None,
               sortdata: bool = True):
    if yfunc is not None:
      y = map(yfunc, y)
      if yerror is not None:
        yerror = map(yfunc, yerror)
    self.x = x
    self.y = y
    self.xerror = xerror
    self.yerror = yerror

    if sortdata:
      self.x, self.y, self.xerror, self.yerror = self.__class__.__sort(self.x, self.y, self.xerror, self.yerror)

    self.label = _tpl_curve_label(
        token="addlegendentry", value=label,
        suppress_semicolon=True
    )
    self.tablename = "table%d" % self.__class__.__counter
    self.__class__.__counter += 1
    self.color = _tpl_attribute_variable(token="color", value=color)
    self.opacity = _tpl_attribute_variable(token="opacity", value=opacity)



import setuptools


with open("README.md", "r", encoding="utf-8") as readme:
  long_description = readme.read()


setuptools.setup(
  name="texplotlibx",
  version="0.8.3",
  author="D Hir",
  author_email="d_hir@outlook.com",
  description="A plot library for creating graphics using tikz/pgfplots",
  long_description=long_description,
  long_description_content_type="text/markdown",
  url="https://gitlab.com/d_hir/texplotlib",
  classifiers=[
    "Programming Language :: Python :: 3",
    "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
    "Operating System :: POSIX",
    "Development Status :: 3 - Alpha"
  ],
  package_dir={"": "src"},
  packages=setuptools.find_packages(where="src"),
  python_requires=">=3.6"
)




from .curve_data import _tpl_curve_data
from .curve_generator.curve_generator import _tpl_curve_generator
from .curve_generator.get_generator import _get_generator


class _tpl_curve:
  def __init__(self, *args, **kwargs):
    self.__data = _tpl_curve_data(*args, **kwargs)
    self.__generator = _tpl_curve_generator(self.__data)

  def _set_plottype(self, plottype: str = 'xy'):
    gen_cls = _get_generator(plottype)
    if gen_cls is not None:
      self.__generator = gen_cls(self.__data)

  def _get_addplot_code(self):
    return self.__generator._get_addplot_code()



from .attribute import _tpl_attribute_function, _tpl_attribute


class _tpl_curve_label(_tpl_attribute_function):
  def __str__(self):
    v = self.get_value()
    if v == '_nolegend_':
      return _tpl_attribute.__str__(self)
    else:
      return super().__str__()
